﻿/*
 * tsja.c
 */

#include	<stdio.h>
#include	<ctype.h>
#include	<stdarg.h>
#include	<stdlib.h>
#include	<string.h>

#include	<mecab.h>

#include	"tsja.h"

typedef	struct	{
	int		id;
	char *		name;
	char *		word_class;
	char *		subtype_1;
	char *		subtype_2;
	char *		subtype_3;
} NAME_TABLE;
#define	ID_WORD	30
NAME_TABLE	word_class_id_tbl[] = {
	{ ID_WORD, "word",       "word", NULL, NULL, NULL },
						/* featureには現れない例外 */

	{ 151, "非自立名詞",     "名詞", "非自立", NULL, NULL },
	{ 152, "代名詞",         "名詞", "代名詞", NULL, NULL },
	{ 153, "数",             "名詞", "数", NULL, NULL },
	{ 154, "助数詞",         "名詞", "接尾", "助数詞", NULL },
	{  50, "名詞",           "名詞", NULL, NULL, NULL },

	{ 161, "非自立動詞",     "動詞", "非自立", NULL, NULL },
	{  60, "動詞",           "動詞", NULL, NULL, NULL },

	{ 171, "非自立形容詞",   "形容詞", "非自立", NULL, NULL },
	{  70, "形容詞",         "形容詞", NULL, NULL, NULL },

	{ 181, "助詞類接続副詞", "副詞", "助詞類接続", NULL, NULL },
	{  80, "副詞",           "副詞", NULL, NULL, NULL },

	{  90, "連体詞",         "連体詞", NULL, NULL, NULL },

	{ 211, "助詞",     "助詞",     NULL, NULL, NULL },
	{ 212, "助動詞",   "助動詞",   NULL, NULL, NULL },
	{ 213, "接続詞",   "接続詞",   NULL, NULL, NULL },
	{ 214, "接頭詞",   "接頭詞",   NULL, NULL, NULL },
	{ 215, "感動詞",   "感動詞",   NULL, NULL, NULL },
	{ 221, "記号",     "記号",     NULL, NULL, NULL },
	{ 222, "フィラー", "フィラー", NULL, NULL, NULL },
	{ 223, "その他",   "その他",   NULL, NULL, NULL },
	{ 255, "BOS/EOS",  "BOS/EOS",  NULL, NULL, NULL },
};
		/* idには255以下の一意的な値を割り当てる。*/
#define	IS_STOP_WORD(id)	((id) >= 100)
		/* ストップ・ワードには100以上のidを割り当てる。 */

#define	MAX_DICT_PATH	256
#define	CONF_MECAB_DICT_PATH	"tsja.mecab_dict_path"
static	char	g_dict_path[MAX_DICT_PATH] = "/usr/local/lib/mecab/dic/ipadic";
		/* mecab辞書の場所: postgresql.conf で設定可能 */

#ifdef	PGSQL_EXTENSION

#include	"postgres.h"
#include	"fmgr.h"

#include	"commands/defrem.h"
#include	"lib/stringinfo.h"
#include	"nodes/pg_list.h"
#include	"tsearch/ts_public.h"
#include	"utils/guc.h"

PG_MODULE_MAGIC;


extern void PGDLLEXPORT _PG_init(void);
extern void PGDLLEXPORT _PG_fini(void);
extern Datum PGDLLEXPORT tsja_start(PG_FUNCTION_ARGS);
extern Datum PGDLLEXPORT tsja_gettoken(PG_FUNCTION_ARGS);
extern Datum PGDLLEXPORT tsja_end(PG_FUNCTION_ARGS);
extern Datum PGDLLEXPORT tsja_init(PG_FUNCTION_ARGS);
extern Datum PGDLLEXPORT tsja_lextypes(PG_FUNCTION_ARGS);
extern Datum PGDLLEXPORT tsja_lexize(PG_FUNCTION_ARGS);

#else	/* PGSQL_EXTENSION */

#define	palloc(sz)	malloc(sz)
#define	palloc0(sz)	malloc(sz)
#define	pfree(ptr)	free(ptr)
#define	pstrdup(in)	strdup(in)

#endif	/* PGSQL_EXTENSION */

/* ---------------------------------------------------------------------
*/
int	info_out(const char * fmt, ...)
{
	va_list		ap;
	char		buf[256];
	int		n = 0;

	va_start(ap, fmt);
	n = vsnprintf(buf, 256, fmt, ap);
#ifdef	PGSQL_EXTENSION
	elog(INFO, "%s", buf);
#else
	fprintf(stderr, "%s\n", buf);
#endif
	va_end(ap);
	return n;
}

#ifdef	PGSQL_EXTENSION

/* =====================================================================
*/
void	_PG_init(void)
{
	const char *	conf_val;

	if ((conf_val = GetConfigOption(CONF_MECAB_DICT_PATH, true, false)) != NULL) {
		strncpy(g_dict_path, conf_val, MAX_DICT_PATH);
	}
	mecab_initialize();
}

void	_PG_fini(void)
{
	mecab_finalize();
}

/* =====================================================================
*/
PG_FUNCTION_INFO_V1(tsja_start);
Datum	tsja_start(PG_FUNCTION_ARGS)
{
	char *		source = (char *) PG_GETARG_POINTER(0);
	int		len = PG_GETARG_INT32(1);
	tsja_parser *	ja_parser;

	ja_parser = start_ja_parser2(source, len);

	PG_RETURN_POINTER(ja_parser);
}

/* =====================================================================
*/
PG_FUNCTION_INFO_V1(tsja_gettoken);
Datum	tsja_gettoken(PG_FUNCTION_ARGS)
{
	tsja_parser *	parser = (tsja_parser *) PG_GETARG_POINTER(0);
	const char **	t = (const char **) PG_GETARG_POINTER(1);
	int *		tlen  = (int *) PG_GETARG_POINTER(2);
	int		lextype;

	lextype = ja_gettoken(parser, (char **) t, tlen);

	PG_RETURN_INT32(lextype);
}

/* =====================================================================
*/
PG_FUNCTION_INFO_V1(tsja_end);
Datum	tsja_end(PG_FUNCTION_ARGS)
{
	tsja_parser *parser = (tsja_parser *) PG_GETARG_POINTER(0);

	end_ja_parser(parser);

	PG_RETURN_VOID();
}

/* =====================================================================
** 参考 (src/backend/tsearch/wparser_def.c)
*/
PG_FUNCTION_INFO_V1(tsja_lextypes);
Datum	tsja_lextypes(PG_FUNCTION_ARGS)
{
	size_t		num_entries = sizeof(word_class_id_tbl) / sizeof(NAME_TABLE);
	LexDescr *	descr = (LexDescr *) palloc0(sizeof(LexDescr) * (num_entries + 1));
	int		i;

	for (i = 0; i < num_entries; i++) {
		descr[i].lexid = word_class_id_tbl[i].id;
		descr[i].alias = pstrdup(word_class_id_tbl[i].name);
		descr[i].descr = pstrdup(word_class_id_tbl[i].name);
	}

	descr[num_entries].lexid = 0;

	PG_RETURN_POINTER(descr);
}

/* =====================================================================
*/
PG_FUNCTION_INFO_V1(tsja_init);
Datum	tsja_init(PG_FUNCTION_ARGS)
{
	List *			dictoptions = (List *) PG_GETARG_POINTER(0);
	tsja_dictionary *	ja_dictionary;
	ListCell *		lc;
	char *			opt_value = "";

	foreach(lc, dictoptions) {
		DefElem *	defel = (DefElem *) lfirst(lc);
		if (pg_strcasecmp(defel->defname, "OptionSample") == 0) {
			opt_value = defGetString(defel);
		}
	}			/*
				** この方法で、
				**	CREATE TEXT SEARCH DICTIONARY
				** に指定されたオプションを取得できる。
				*/

	ja_dictionary = (tsja_dictionary *) palloc0(sizeof(tsja_dictionary));
	ja_dictionary->magic = DICTIONARY_MAGIC;

	PG_RETURN_POINTER(ja_dictionary);
}

/* =====================================================================
** テキスト検索テンプレートの LEXIZE として登録する函数。
** 語彙を正規化する。
** ---- 実際には、該当する単語のみ取り出して正規化を試みるのではなく、
** ---- 文脈の中で何に当たるかを考慮して正規化しなければならない。
*/
PG_FUNCTION_INFO_V1(tsja_lexize);
Datum	tsja_lexize(PG_FUNCTION_ARGS)
{
	/* tsja_dictionary * ja_dictionary = (tsja_dictionary *) PG_GETARG_POINTER(0); */
	const char *	t = (char *) PG_GETARG_POINTER(1);
	int		tlen = PG_GETARG_INT32(2);
	TSLexeme *	res = palloc0(sizeof(TSLexeme) * 2);

	char *		basic_form;
	int		is_stop_word = 0;

	basic_form = ja_lexize(t, tlen, &is_stop_word);
	if (basic_form == NULL) {
		basic_form = to_basic_form2((char *) t, tlen, &is_stop_word);
	}
	if (basic_form == NULL) {
		basic_form = pstrdup("UNKNOWN");
	}
	if (! is_stop_word) {
		res[0].lexeme = basic_form;
	}

	PG_RETURN_POINTER(res);
}

/* =====================================================================
*/
PG_FUNCTION_INFO_V1(pg_to_ruby);
Datum	pg_to_ruby(PG_FUNCTION_ARGS)
{
	text *	t = PG_GETARG_TEXT_P(0);
	char *	src = VARDATA(t);
	char *	dest = to_ruby(src);
	size_t	len = strlen(dest);

	text *	result = (text *) palloc0(VARHDRSZ + len);
	SET_VARSIZE(result, VARHDRSZ + len);
	memcpy(VARDATA(result), dest, len);

	pfree(dest);

	PG_RETURN_TEXT_P(result);
}

#endif	/* PGSQL_EXTENSION */




/* =====================================================================
*/
static	mecab_t *	g_tagger = NULL;
static	mecab_node_t *	g_curr_node = NULL;

/* =====================================================================
*/
void	set_mecab_dict_path(const char * dict_path)
{
	strncpy(g_dict_path, dict_path, MAX_DICT_PATH);
}

/* =====================================================================
*/
void	mecab_initialize(void)
{
	char		mecab_arg[32 + MAX_DICT_PATH];

	if (g_tagger == NULL) {
		snprintf(mecab_arg, 32 + MAX_DICT_PATH,
				"-Ochasen -d%s", g_dict_path);
		g_tagger = mecab_new2(mecab_arg);
	}
}

/* =====================================================================
*/
void	mecab_finalize(void)
{
	if (g_tagger != NULL) {
		mecab_destroy(g_tagger);
		g_tagger = NULL;
	}
}

/* =====================================================================
*/
typedef	int	boolean;
#define	TRUE	1
#define	FALSE	0

#define IS_BOS_EOS(stat)	(((stat) == MECAB_BOS_NODE) || ((stat) == MECAB_EOS_NODE))

#define	MECAB_FEATURE_WORD_CLASS	0
#define	MECAB_FEATURE_SUBTYPE_1		1
#define	MECAB_FEATURE_SUBTYPE_2		2
#define	MECAB_FEATURE_SUBTYPE_3		3
#define	MECAB_FEATURE_CONJ_TYPE		4
#define	MECAB_FEATURE_CONJUGATION	5
#define	MECAB_FEATURE_BASIC_FORM	6
#define	MECAB_FEATURE_RUBY		7
#define	MECAB_FEATURE_PRONOUNCE		8

/* ---------------------------------------------------------------------
*/
int	to_word_class_id(char * feature)
{
	char *	word_class = get_column_from_csv(feature, MECAB_FEATURE_WORD_CLASS);
	char *	subtype_1 = get_column_from_csv(feature, MECAB_FEATURE_SUBTYPE_1);
	char *	subtype_2 = get_column_from_csv(feature, MECAB_FEATURE_SUBTYPE_2);
	char *	subtype_3 = get_column_from_csv(feature, MECAB_FEATURE_SUBTYPE_3);
	int	id = 0;
	int	i;
	for (i = 0; i < sizeof(word_class_id_tbl) / sizeof(NAME_TABLE); i++) {
		if ((word_class_id_tbl[i].word_class != NULL) &&
		    (strcmp(word_class_id_tbl[i].word_class, word_class) != 0)) {
			continue;
		}
		if ((word_class_id_tbl[i].subtype_1 != NULL) &&
		    (strcmp(word_class_id_tbl[i].subtype_1, subtype_1) != 0)) {
			continue;
		}
		if ((word_class_id_tbl[i].subtype_2 != NULL) &&
		    (strcmp(word_class_id_tbl[i].subtype_2, subtype_2) != 0)) {
			continue;
		}
		if ((word_class_id_tbl[i].subtype_3 != NULL) &&
		    (strcmp(word_class_id_tbl[i].subtype_3, subtype_3) != 0)) {
			continue;
		}
		id = word_class_id_tbl[i].id;
		break;
	}
	return id;
}

/* ---------------------------------------------------------------------
*/
int	is_all_ascii(char * str, size_t len)
{
	int	i;
	for (i = 0; i < len; i++) {
		if (! isascii(str[i])) {
			return 0;
		}
	}
	return 1;
}

/* =====================================================================
*/
tsja_parser *	start_ja_parser2(char * str, size_t len)
{
	tsja_parser *	ja_parser = (tsja_parser *) palloc(sizeof(tsja_parser));
	mecab_initialize();
	ja_parser->magic = PARSER_MAGIC;
	ja_parser->tagger = g_tagger;
	ja_parser->node = (mecab_node_t *) mecab_sparse_tonode2(g_tagger, str, len);
	g_curr_node = ja_parser->node;
	return ja_parser;
}

/* =====================================================================
*/
int	ja_gettoken(tsja_parser * ja_parser, char **token, int *len)
{
	const mecab_node_t *	node;

	if (ja_parser->node == NULL) {
		return 0;
	}

	node = ja_parser->node = ja_parser->node->next;
	if (node == NULL) {
		return 0;
	}
	if (IS_BOS_EOS(node->stat)) {
		return 0;
	}

	*token = (char *) node->surface;
	*len = node->length;

	if ((*len > 0) && (is_all_ascii(*token, *len))) {
		return ID_WORD;
	}
	return to_word_class_id((char *) node->feature);
}

/* =====================================================================
*/
char *	ja_lexize(const char * t, int tlen, int *is_stop_word)
{
	int	word_class_id;

	while (g_curr_node != NULL) {
		if ((g_curr_node->length == tlen) &&
		    (memcmp(g_curr_node->surface, t, tlen) == 0)) {
			break;
		}
		else {
			g_curr_node = g_curr_node->next;
		}
	}
	if (g_curr_node == NULL) {
		return NULL;
	}
	word_class_id = to_word_class_id((char *) g_curr_node->feature);
	*is_stop_word = 0;
	if (IS_STOP_WORD(word_class_id)) {
		*is_stop_word = 1;
	}
	return get_column_from_csv((char *) g_curr_node->feature, MECAB_FEATURE_BASIC_FORM);
}

/* =====================================================================
*/
void	end_ja_parser(tsja_parser * parser)
{
	/* mecab_finalize(); */
		/*
		** ここでmecab_destroy()してはならない。
		** この後でja_lexize()を呼び出し、基本形を取り出すなど
		** することがあるため。
		*/
	pfree(parser);
	parser = NULL;
}

/* =====================================================================
** 単語を正規化: strの先頭の字句について、基本形を返す。
*/
char *	to_basic_form2(char * str, size_t len, int *is_stop_word)
{
	const mecab_node_t *	node;
	int	word_class_id;

	mecab_initialize();

	for (node = mecab_sparse_tonode2(g_tagger, str, len); node != NULL; node = node->next) {
		if (IS_BOS_EOS(node->stat)) {
			continue;
		}
		word_class_id = to_word_class_id((char *) node->feature);
		*is_stop_word = 0;
		if (IS_STOP_WORD(word_class_id)) {
			*is_stop_word = 1;
		}
		return get_column_from_csv((char *) node->feature, MECAB_FEATURE_BASIC_FORM);
	}
	return NULL;
}

/* =====================================================================
*/
char *	to_ruby(char * str)
{
	const mecab_node_t *	node;
	char *		ruby;
	char *		result;
	char *		prev_result;
	size_t		len;

	result = palloc(1);
	strcpy(result, "");
	prev_result = result;

	for (node = mecab_sparse_tonode(g_tagger, str); node != NULL; node = node->next) {
		if (IS_BOS_EOS(node->stat)) {
			continue;
		}
		ruby = get_column_from_csv((char *) node->feature, MECAB_FEATURE_RUBY);
		len = strlen(result) + strlen(ruby);
		prev_result = result;
		result = palloc(len + 1);
		strcpy(result, prev_result);
		strcat(result, ruby);

		pfree(prev_result);
		pfree(ruby);
	}
	return result;
}

/* =====================================================================
** 戻り値: nth番目の列。用済み後、free() すること。
*/
char *	get_column_from_csv(char * str, int nth)
{
	int	i;
	char *	begin = str;
	char *	end;
	size_t	len;
	char *	result;

	for (i = 0; i < nth; i++) {
		begin = strchr(begin, ',');
		if (begin == NULL) {
			return NULL;
		}
		begin += 1;
	}
	end = strchr(begin, ',');
	if (end == NULL) {
		len = strlen(begin);
	}
	else {
		len = end - begin;
	}
	result = palloc(len + 1);
	memcpy(result, begin, len);
	result[len] = '\0';
	return result;
}

